#################################################################
#################################################################
##################### modcone_genus_2 ###########################
#################################################################
#################################################################
#  A way to provide empirical evidences for the polyhedrality   #
#   of the modular cone of weight k, whenever k is congruent    #
#                     to 2 modulo 4                             #
#################################################################
#################################################################

import time
import sys
Z=IntegerRing()

#################################################################
######### COMPUTATION OF MODULAR CONES in genus 2 ###############
#################################################################

def compute_and_save_bases(k,prec=30):
    """
    we compute all the basis with weights from 16 to 2*limit.
    """
    if k<9:
        print("This is a trivial case, since the only Siegel cusp form is 0.")
        return
    print('Computation of basis of S^'+str(k)+'_2(\QQ) \oplus N^'+str(k)+'_2(\QQ) in progress...')
    a='basis'+str(k)+'_prec_'+str(prec)
    basis=KlingenEisensteinAndCuspForms(k,prec)
    basis.save_basis_as_binary(a)
    print('Basis has been saved under the name: '+a)
    print('Computation of the Siegel Eisenstein series E^'+str(k)+'_2 in progress...')
    b='eis'+str(k)+'_prec_'+str(prec)
    eis=eisenstein_series_degree2(k,prec)
    eis.save_as_binary(b)
    print('Siegel Eisenstein series has been saved under the name: '+b)
    return

def compute_and_save_coefficient_extraction_functionals(k,prec=30):
    """
    with precision=prec, this function save in coeff_TOTXX the set of (different!) coefficients, remembering also list of determinants and division nperdet
    """
    if k<9:
        print("This is a trivial case, since the only Siegel cusp form is 0.")
        return
    print("CASE WEIGHT=%d") % k
    #here we download files
    print("I am downloading files...")
    a='eis'+str(k)+"_prec_"+str(prec)
    eis=load(a)
    eis=eis['fc_dct']
    a='basis'+str(k)+"_prec_"+str(prec)
    v=KlingenEisensteinAndCuspForms(k,prec)
    v.load_basis_from(a)
    basisTOT=v.basis()
    print("I'm computing the set of c_T under the previous basis with precision"+str(prec))
    #here we get the set of matrices which are positive definite and half-integral
    quickmat=[]
    quicktemp=x10_with_prec(prec)
    quicktemp=quicktemp.sorted_list()
    for i in range(len(quicktemp)):
        quickmat=quickmat+[quicktemp[i][0]]
    ordmat=[]
    d={}
    for i in quickmat:
        a=coerce(QQ,i[0]); b=coerce(QQ,i[1]); c=coerce(QQ,i[2]); #need to force these to be integral
        M=matrix(QQ,[[a,b/2],[b/2,c]]);
        deter=det(M);
        d[deter]=[]
    for T in quickmat:
        a=coerce(QQ,T[0]); b=coerce(QQ,T[1]); c=coerce(QQ,T[2]); # need to force these to be integral
        M=matrix(QQ,[[a,b/2],[b/2,c]]);
        deter=det(M);
        d[deter]=d[deter]+[T]
    temp=sorted(d.keys())
    for j in temp:
        ordmat=ordmat+d[j]
    quickmat=ordmat # here we order the matrices w.r.t. the determinant
    detdict=d # This is the dictionary of matrices in quickmat divided by determinant
    determinants=sorted(detdict.keys()) # list of determinants
    basisTOT=basisTOT+[eis] # here we add the Siegel-Eisenstein series

    # Here we extract the points c_T, already ordered by det(T)
    points=extraction(basisTOT,quickmat) 
    quickpoints=[]
    if points==[]:
        print("The precision in not enought. ABORT!")
        return
    quickpoints=quickpoints+[points[0]]
    quicklen=1
    lenpoints=len(points)
    for j in range(1,len(points)):
        b=False
        for h in range(quicklen):
            if (quickpoints[h]==points[j]):
                b=True
        if not b:
            quickpoints=quickpoints+[points[j]]
            quicklen=quicklen+1
        if b:
            # delete from detdict the corresponding matrix!
            T=quickmat[j]
            a=coerce(QQ,T[0]); b=coerce(QQ,T[1]); c=coerce(QQ,T[2]); # need to force these to ge integral
            M=matrix(QQ,[[a,b/2],[b/2,c]]);
            tdet=det(M)
            (detdict[tdet]).remove(T)
    nperdet=[] # useful to truncate quickmat per determinants
    nperdet=nperdet+[len(detdict[determinants[0]])]
    for g in range(1,len(determinants)):
        nperdet=nperdet+[nperdet[g-1]+len(detdict[determinants[g]])]
    a=[detdict,nperdet,quickpoints]
    b="coeff_TOT"+str(k)+"_prec_"+str(prec)
    save(a,b)
    print('The functionals c_T has been saved under the name: ' + b)
    return

def compute_and_save_cone(k,prec=30):
    """
    We compute the cone C for weight k uploading the files previously saved.
    We check that after adding the functionals of low determinants, the cone is stable.
    """
    if k<9:
        print("This is a trivial case, since the only Siegel cusp form is 0.")
        return
    #we fix tail=1
    print("CASE OF WEIGHT k=%d.") % k
    if mod(k,4)!=2:
        print("ATTENTION! The weight is NOT congruent to 2 mod 4.")
        
    #here we download files
    a='coeff_TOT'
    a=a+str(k)+"_prec_"+str(prec)
    coeff_TOT=load(a)
    detdict=coeff_TOT[0]
    determinants=sorted(detdict.keys())
    nperdet=coeff_TOT[1]
    quickpoints=coeff_TOT[2]

    #Let's start with Polyhedron!
    w=quickpoints[0:nperdet[0]] # we take c_T associated to T=(1,1,1)
    tempP=Polyhedron(rays=w)
    print("With just the first determinant, we get: "+str(tempP)+".")
    print("Now we begin to increase the determinants.")
    lencT=len(quickpoints[0])
    y=1
    a="coneC"+str(k)+"_prec_"+str(prec)
    yesno=1
    for f in range(1,len(determinants)):
        y=y+1
        r=tempP.rays()
        add=[]
        for j in range(len(r)):
            rel=zero_vector(QQ,lencT)
            for jj in range(lencT):
                rel[jj]=r[j][jj]
            add=add+[rel]
        w=quickpoints[nperdet[f-1]:nperdet[f]]
        w=add+w
        P=Polyhedron(rays=w)
        #print "Is the polyhedron == to the previous one? We take c_T with detT<=%s" % determinants[f]
        if P==tempP:
            if yesno==1:
                print("Same cone as above.")
                print("Number of determinants considered: "+str(y)+".")
                yesno=0
            else:
                sys.stdout.write('\x1b[1A')
                sys.stdout.write('\x1b[2K')
                print("Number of determinants considered: "+str(y)+".")
        else:
            yesno=1
            tempP=P
            print("Different cone! We get: "+str(P)+".")
            P.save(a)
            print("Number of determinants considered: "+str(y)+".")
    return P

def c_T(k,n,r,m,prec=30):
    """
    This computes the functional c_T under usual basis:
    cusp forms --- Klingen Eisenstein --- Siegel Eisenstein.
    """
    if k<9:
        print("This is a trivial case, since the only Siegel cusp form is 0.")
        return
    mat=(n,r,m)
    KC=KlingenEisensteinAndCuspForms(k,30)
    KC.load_basis_from("basis"+str(k)+"_prec_"+str(prec))
    BasisKC=KC.basis()
    eis=load("eis"+str(k)+"_prec_"+str(prec))
    eis=eis["fc_dct"]
    cT=[]
    for f in BasisKC:
        cT=cT+[f[mat]]
    cT=cT+[eis[mat]]
    return cT
    
#################################################################
######### ACCUMULATION RAYS AND CONES in genus 2 ################
#################################################################

def V_m(k,m,prec=30):
    """
    This function compute the points V_m as in:
        Zuffetti - "Cones of special cycles of codimension 2 on orthogonal Shimura varieties".
    Notice: the choice of the basis here is:
    cusp forms --- Klingen Eisenstein --- Siegel Eisenstein.
    Moreover: when we take a basis of elliptic cusp forms in SageMath, these are Hecke forms, in particular are normalized.
    But the package Stakemori for degree 2 modular forms gets Klingen-Eis series via the generators of the ring of Siegel modular forms.
    So you loose the connection with cusp forms of lower degree.
    The point (*) is the part in the program where this remark is crucial.
    """
    if k<9:
        print("This is a trivial case, since the only Siegel cusp form is 0.")
        return
    KC=KlingenEisensteinAndCuspForms(k,prec)
    a="basis"+str(k)+"_prec_"+str(prec)
    KC.load_basis_from(a)
    BasisKC=KC.basis()
    numcusp=KC.dimension_of_cuspforms()
    numkling=KC.dimension()-numcusp
    Vm=[0]*(numcusp)
    for i in range(numkling):
        #(*)
        temp_f=BasisKC[numcusp+i]
        Vm=Vm+[zeta(1-k)/2*alpha_m(1,temp_f,m,k)]
    Vm=Vm+[1]
    return Vm

def accumulation_cone_Hecke(k,max=100):
    """
    This function compute the accumulation cone over the basis of Hecke forms of M^k_1.
        In fact, we forget about the first cooedinates w.r.t. the weight k Siegel cusp forms.
        It is the cone constructed as in "gen1_compute_and_save_cone_Hecke(k,max=100)",
        but using the points V_n instead of c_n.
    """
    if k<15 and not k == 12:
        print("This is a trivial case, since the only elliptic cusp form is 0.")
        return
    Mk=ModularForms(1,k)
    Sk=Mk.cuspidal_subspace()
    v=Sk.newforms(names='a')
    f=v[0]
    nfield=f.base_ring()
    emb=nfield.embeddings(AA) # embeddings into AA=Real Algebraic Numbers
    coefcusp=list(f.q_expansion(max+2))
    listbasis=[]
    for embedding in emb:
        aux=[]
        for n in range(max+1):
            aux = aux + [embedding(coefcusp[n])]
        listbasis=listbasis + [aux]
    listVm=[]
    for m in range(1,max+1):
        Vm=[]
        for i in listbasis:
            Vm=Vm+[zeta(1-k)/2*alpha_m_gen_1(1,i,m,k)]
        Vm=Vm+[1]
        listVm=listVm+[Vm]
    P=Polyhedron(rays=listVm)
    a='coneC'+str(k)+'_max_'+str(max)+'_Hecke_with_Vm'
    P.save(a)
    return P

#################################################################
########### SOME INTERESTING SUBCONES in genus 2 #################
#################################################################

def subcone_for_V_m(k,m,prec=30):
    """
    For every m, we construct the cone that we conjectured should contain Vm as internal point, see [Zuffetti, Conjecture 3]
    For m square-free, it is proved theoretically in a Lemma in my notes.
    """
    print("CASE OF WEIGHT k="+str(k)+" AND INDEX m="+str(m)+".")
    Vm=V_m(k,m,prec)
    sqdiv=[]
    for i in range(2,m+1):
        if Z(i*i).divides(m):
            sqdiv=sqdiv+[i]
    print("SQUARE-DIVISORS =!1 OF "+str(m)+" are: "+str(sqdiv))
    # We first contruct the cone associated to matrices where n <= m. We will enlarge it afterwards.
    v=[]
    for n in range(1,m+1): # Here I modified m with m+1... (date: 2022)
        rset=[]
        for r in range(1,ceil(sqrt(4*n*m))):
            b=True
            if not 4*n*m-r*r>0:
                b=False
            else:
                for t in sqdiv:
                    if Z(t).divides(r):
                        b=False
            if b:
                rset=rset+[r]
        print("Case n="+str(n)+". ADMISSIBLE ENTRIES FOR r are: "+str(rset)+".")
        for r in rset:
            v=v+[c_T(k,n,r,m,prec)]
    print("I am computing the cone gen. by the previous c_T.")
    P=Polyhedron(rays=v)
    print("We run enlarge the cone step by step, for all n>"+str(m)+".")
    b=True
    auxb=True
    for n in range(m+1,prec+1):
        rset=[]
        for r in range(1,ceil(sqrt(4*n*m))):
            bb=True
            for t in sqdiv:
                if Z(t).divides(r):
                    bb=False
            if bb:
                if 4*n*m-r*r>0:
                    rset=rset+[r]
        if b:
            print("Case n="+str(n)+". ADMISSIBLE ENTRIES FOR r are: "+str(rset)+".")
        prays=P.rays_list()
        for r in rset:
            v=v+[c_T(k,n,r,m,prec)]
        tempP=Polyhedron(rays=prays+v)
        if b:
            if tempP.relative_interior_contains(Vm):
                print("The ray given by Vm is INTERNAL!")
                b=False
            elif tempP.contains(Vm):
                print("The subcone contains Vm, and is of dimension "+str(tempP.dimension())+".")
            else:
                print("Vm is NOT contained in the subcone.")
        else:
            if tempP==P:
                if auxb==False:
                    sys.stdout.write('\x1b[1A')
                    sys.stdout.write('\x1b[2K')
                print("Case n="+str(n)+". Same cone as above.")
                auxb=False
            else:
                auxb=True
                if tempP.relative_interior_contains(Vm):
                    print("Different cone. But the ray given by Vm is still INTERNAL!")
                else:
                    print("Vm is not internal anymore...")
                    b=True
        P=tempP
    P.save("subconeR"+str(k)+"_prec_"+str(prec)+"_for_V_"+str(m))
    return P

def subcone_for_V_m_enlarged(k,m,prec=30):
    """
    Here, we consider the cone with all the possible c_T where the bottom-right entry of T is m, even not reduced.
    """
    print("CASE OF WEIGHT k="+str(k)+" AND INDEX m="+str(m)+".")
    v=[]
    print("Case n=1")
    for n in range(1,prec+1):
        for r in range(0,4*n*m+1):
            if 4*m*n-r*r>0:
                b=True
                vtemp=c_T(k,n,r,m)
                if not vtemp in v:
                    v=v+[vtemp]
        P=Polyhedron(rays=v)
        sys.stdout.write('\x1b[1A')
        sys.stdout.write('\x1b[2K')
        print("Case n="+str(n))
        v=P.rays_list()
        #save(P,"subcone_ALL_inside_k="+str(k)+"_m="+str(m)+"_and_n_until_"+str(n)+"_with_all_r")
    P.save("subconeR"+str(k)+"_prec_"+str(prec)+"_for_V_"+str(m)+"_enlarged")
    return P

#################################################################
######### COMPUTATION OF MODULAR CONES in genus 1 ###############
#################################################################

def gen1_compute_and_save_cone(k,max=100):
    """
    Computation of the genus 1 modular cone, using the basis given by SageMath with CuspForms(1,k).basis().
    Here, we compute in one shot the full cone. Then we check whether the accumulation ray given by \omega
        (or -\omega, depending on k) is INTERNAL inside the cone or not.
    """
    print('CASE OF WEIGHT k='+str(k)+':')
    Mk=ModularForms(1,k)
    basis=Mk.basis()
    listbasis=[]
    for b in basis:
        listbasis=listbasis + [list(b.q_expansion(max+2))]
    coefextr=[]
    for n in range(1,max+1):
        aux=[]
        for j in listbasis:
            aux=aux+[j[n]]
        coefextr=coefextr + [aux]
    P=Polyhedron(rays=coefextr)
    print(P)
    a='coneC'+str(k)+'_max_'+str(max)
    P.save(a)
    # We construct \omega or -\omega, depending whether k is cong to 2 or 0 mod 4.
    w=[0]*len(basis)
    if mod(k,4)==2:
        w[len(basis)-1]=-1
    else:
        w[len(basis)-1]=1
    # We check whether \omega or -\omega is contained in the interior or in the boundary or in the complement of the cone.
    if P.interior_contains(w):
        print('The cone contains the dual of E^k_1, and it is IN ITS INTERIOR!')
    else:
        if P.contains(w):
            print('The cone contains the dual of E^k_1, but ON THE BOUNDARY!')
        else:
            print('The cone does NOT contain dual of E^k_1.')
    return P
    
# Every genus 1 modular cone seems to be as it should (polyhedral, with omega or -omega internal)
#            for every even k <= 134, if you use max=130 .

def gen1_compute_and_save_cone_Hecke(k,max=100):
    """
    This is like "gen1_compute_and_save_cone", but we write the functionals c_n w.r.t. a basis over \RR
        made of Hecke forms. More precisely:
            ...newforms... + (normalized) Eisenstein series
    """
    print('CASE OF WEIGHT k='+str(k)+':')
    Mk=ModularForms(1,k)
    basis=Mk.basis()
    eis=basis[len(basis)-1]
    Sk=Mk.cuspidal_subspace()
    v=Sk.newforms(names='a')
    f=v[0]
    nfield=f.base_ring()
    emb=nfield.embeddings(AA) # embeddings into AA=Real Algebraic Numbers
    coefcusp=list(f.q_expansion(max+2))
    coefeis=list(eis.q_expansion(max+2))
    coefextr=[]
    for n in range(1,max+1):
        aux=[]
        for ff in emb: # here we add the coeff given by the Hecke forms
            aux=aux+[ff(coefcusp[n])]
        aux=aux+[coefeis[n]]
        coefextr=coefextr + [aux]
    P=Polyhedron(rays=coefextr,base_ring=AA)
    print(P)
    a='coneC'+str(k)+'_max_'+str(max)+'_Hecke'
    P.save(a)
    # We construct \omega or -\omega, depending whether k is cong to 2 or 0 mod 4.
    w=[0]*len(basis)
    if mod(k,4)==2:
        w[len(basis)-1]=-1
    else:
        w[len(basis)-1]=1
    # We check whether \omega or -\omega is contained in the interior or in the boundary or in the complement of the cone.
    if P.interior_contains(w):
        print('The cone contains the dual of E^k_1, and it is IN ITS INTERIOR!')
    else:
        if P.contains(w):
            print('The cone contains the dual of E^k_1, but ON THE BOUNDARY!')
        else:
            print('The cone does NOT contain the dual of E^k_1.')
    return P

def gen1_check_if_internal_Hecke(k,m,max=100,accumulation=False):
    """
    This function tells you if all functionals c_n, with n>=m, are contained in the *interior* of the modular cone of genus 1 and weight k.
        It uses an upper bound, if the answer is "Yes", then it *proves* that the extremal rays are generated by the first m functionals.
    If "accumulation" is set to "True", then it does the same thing, but:
        -- with V_n in place of c_n.
        -- with the *accumulation cone* of the genus 2 modular cone in place of the genus 1 moduar cone.
    If the answer is "Yes", then it *proves* that the extremal rays  of the accumulation cone are generated by the first m points V_n.
    """
    if k<15 and not k == 12:
        print("This is a trivial case, since the only elliptic cusp form is 0.")
        return
    if m>max:
        print("Invalid input! m must be smaller than max. Abort.")
        return
    S=CuspForms(1,k)
    ell=S.dimension()
    #In what follows, we check whether each of the hyperplain defining the boundary of the saved cone has distance more that r from P_infty. i.e. if the cone cotain the ball in its *interior*.
    if accumulation==False:
        r = sqrt(ell * zeta(1-k)^2 / m^(k-2)) # This is the ray of the ball centered in P_\infty containing all functionals c_n/c_n(E^k_1) with n>=m.
        P=load('coneC'+str(k)+'_max_'+str(max)+"_Hecke")
        eqhyp=P.inequalities_list() # coeff of equation plane cutting boundary faces of the cone. First entry is the constant term.
    elif accumulation==True:
        r = sqrt(ell * zeta(1-k)^2 / m^(k-3))
        P=load('coneC'+str(k)+'_max_'+str(max)+"_Hecke_with_Vm")
        eqhyp=P.inequalities_list() # coeff of equation plane cutting boundary faces of the cone. First entry is the constant term.
    else:
        print("Wrong input for the entry ***accumulation***. Abort.")
        return
    for eq in eqhyp:
        norm=0
        for j in range(1,len(eq)): # We discard the first entry. Anyway: it is zero...
            norm = norm + eq[j]^2
        aux=abs(eq[len(eq)-1])/sqrt(norm) # This is the distance between P_\infty and the plane of equation eq. We use the standard formula for distance point hyperplane.
        # Note: it does not really matter if you take P_\infty or -P_\infty. In fact, the hyperplanes above pass always through the origin!
        if aux < r:
            print('No, the value m='+str(m)+' might be too small. Please, increase it!')
            return
    if accumulation==False:
        print('Yes, all functionals c_n, where n >= '+str(m)+', are contained in the *interior* of the saved cone!')
    elif accumulation==True:
        print('Yes, all points V_n, where n >= '+str(m)+', are contained in the *interior* of the saved cone!')
    return

def gen1_c_n(k,n):
    """
    Compute the functional c_n over the basis of M^k_1 given by
        cusps + E^k_1.
    """
    Mk=ModularForms(1,k)
    basis=Mk.basis()
    fun=[]
    for j in basis:
        fun=fun+[j.q_expansion(n+1)[n]]
    return fun

def gen1_compute_cone_step_by_step(k,max=100):
    """
    We add the coefficient extraction functionals c_n one by one, and we look at what happens to the cone.
    We write the functionals c_n over the basis given by SageMath with ModularForms(1,k).basis()
    This function does NOT save the resulting cone.
    """
    print('CASE OF WEIGHT k='+str(k)+':')
    Mk=ModularForms(1,k)
    basis=Mk.basis()
    listbasis=[]
    for b in basis:
        listbasis=listbasis + [list(b.q_expansion(max+2))]
    coefextr=[]
    for n in range(max+1):
        aux=[]
        for j in listbasis:
            aux=aux+[j[n]]
        coefextr=coefextr + [aux]
    
    gen=[coefextr[1]]
    lin=[]
    P=Polyhedron(rays=gen)
    yesno=1
    for n in range(2,max+1):
        gen=gen + [coefextr[n]]
        tempP=Polyhedron(rays=gen,lines=lin)
        if tempP==P:
            if yesno==1:
                print('n <= '+str(n)+' ---> Same cone.')
                yesno=0
            else:
                sys.stdout.write('\x1b[1A')
                sys.stdout.write('\x1b[2K')
                print('n <= '+str(n)+' ---> Same cone.')
        else:
            yesno=1
            print('n <= '+str(n)+' ---> Different! '+str(tempP))
        P=tempP
        gen=P.rays_list()
        lin=P.lines_list()
    # We construct \omega or -\omega, depending whether k is cong to 2 or 0 mod 4.
    w=[0]*len(basis)
    if mod(k,4)==2:
        w[len(basis)-1]=-1
    else:
        w[len(basis)-1]=1
    # We check whether \omega or -\omega is contained in the interior or in the boundary or in the complement of the cone.
    if P.interior_contains(w):
        print('The cone contains the accumulation ray, and it is IN ITS INTERIOR!')
    else:
        if P.contains(w):
            print('The cone contains the accumulation ray, but ON THE BOUNDARY!')
        else:
            print('The cone does NOT contain the accumulation ray!')
    return

#################################################################
################# AUXILIARY FUNCTIONS ###########################
#################################################################   
    
def gen1_Vm(k,m):
    """
    This function compute the points V_m but projected in genus 1.
        Zuffetti - "Cones of special cycles of codimension 2 on orthogonal Shimura varieties".
    Notice: the choice of the basis here is:
    cusp forms ---  Eisenstein
    """
    Mk=ModularForms(1,k)
    Sk=Mk.cuspidal_subspace()
    basis=Sk.basis()
    listbasis=[]
    for b in basis:
        listbasis=listbasis + [list(b.q_expansion(m+2))]
    aux=[]
    for i in listbasis:
        aux=aux+[-zeta(1-k)/2*alpha_m_gen_1(1,i,m,k)]
    aux=aux+[-1]
    return aux
    
def gen1_Vm_Hecke(k,m):
    """
    This function is as "gen1_Vm", but we choose instead a basis of Hecke forms.
        In particular, the functionals are vectors with entries in AA, not animore in QQ.
    """
    Mk=ModularForms(1,k)
    basis=Mk.basis()
    eis=basis[len(basis)-1]
    Sk=Mk.cuspidal_subspace()
    v=Sk.newforms(names='a')
    f=v[0]
    nfield=f.base_ring()
    emb=nfield.embeddings(AA) # embeddings into AA=Real Algebraic Numbers
    coefcusp=list(f.q_expansion(m+2))
    listbasis=[]
    for ff in emb: # here we add the coeff given by the Hecke forms
        listaux=[]
        for n in coefcusp:
            listaux=listaux+[ff(n)]
        listbasis=listbasis+[listaux]
    aux=[]
    for i in listbasis:
        aux=aux+[-zeta(1-k)/2*alpha_m_gen_1(1,i,m,k)]
    aux=aux+[-1]
    return aux    

def extraction(w,matrices):
	points=[]
	l=len(w)
	for T in matrices:
		temp=vector(QQ,l)
		for j in range(l):
			temp[j]=w[j][T]
		points=points+[temp]
	return points

def g_k(k,m):
	"""
	This function computes g_k as in the Böcherer--Das article.
	"""
	g=0
	for i in range(1,ceil(sqrt(m))+1):
		square=i*i
		if Z(square).divides(m):
			g=g+moebius(i)*sigma(m/square,k-1)
	return g

def g_f(f,m):
	"""
	This function computes g_f as in the Böcherer--Das article.
	f = a Klingen--Eisenstein series, we evaluate the coefficients of its associated elliptic cusp form
        (in this way we do not have to deal with normalizations).
	"""
	g=0
	for i in range(1,ceil(sqrt(m))+1):
		square=i*i
		if Z(square).divides(m):
			g=g+moebius(i)*f[m/square,0,0]
	return g

def g_f_gen_1(f,m):
	"""
	This function computes g_f as in the Böcherer--Das article.
	f = a list of coeff of a cusp form, we evaluate the coefficients of its associated elliptic cusp form
        (in this way we do not have to deal with normalizations).
	"""
	g=0
	for i in range(1,ceil(sqrt(m))+1):
		square=i*i
		if Z(square).divides(m):
			g=g+moebius(i)*f[m/square]
	return g

def alpha_m(t,f,m,k):
	"""
	This function computes alpha_m as in the Böcherer--Das article.
	f is a klingen eis series, see "g_f" for more information.
	"""
	alpha=0
	for i in range(1,t+1):
		if Z(i).divides(t):
			alpha=alpha+moebius(t/i)*g_f(f,m/(i*i))/g_k(k,m/(i*i))
	return alpha
	
def alpha_m_gen_1(t,f,m,k):
	"""
	This function computes alpha_m as in the Böcherer--Das article.
	f is cusp form, see "g_f_gen_1" for more information.
	"""
	alpha=0
	for i in range(1,t+1):
		if Z(i).divides(t):
			alpha=alpha+moebius(t/i)*g_f_gen_1(f,m/(i*i))/g_k(k,m/(i*i))
	return alpha
    
    
    
    
    
    